/* /////////////////////////////////////////////////////////////////////////////
 * File:    cstring.c
 *
 * Purpose: The implementation of the cstring API
 *
 * Created: 16th June 1994
 * Updated: 6th October 2005
 *
 * Home:    http://synesis.com.au/software/
 *
 * Copyright 1994-2005, Matthew Wilson and Synesis Software
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer. 
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the names of Matthew Wilson and Synesis Software nor the names of
 *   any contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * ////////////////////////////////////////////////////////////////////////// */


/** \file cstring.c The implementation of the cstring API
 */

/* /////////////////////////////////////////////////////////////////////////////
 * Includes
 */

#include <cstring/cstring.h>

#include <assert.h>
#include <stdlib.h>
#include <string.h>

/* /////////////////////////////////////////////////////////////////////////////
 * Constants & definitions
 */

#define CSTRING_ALLOC_GRANULARITY   (8)

#define CSTRING_OFFSET_SIZE         (16)

/* /////////////////////////////////////////////////////////////////////////////
 * Debugging
 */

#define cstring_assert(expr)    assert(expr)

/* /////////////////////////////////////////////////////////////////////////////
 * Compiler warnings
 */

#if defined(_MSC_VER) && \
	_MSC_VER >= 1310
# pragma warning(disable : 4055)
#endif /* compiler */

/* /////////////////////////////////////////////////////////////////////////////
 * Utility functions
 */

static size_t cstring_strlen_safe_(char const *s)
{
    return (NULL == s) ? 0 : strlen(s);
}

/* /////////////////////////////////////////////////////////////////////////////
 * Allocation functions
 */

#if defined(WIN32)
static void *win32_global_realloc(void *pv, size_t cb);
static void *win32_processheap_realloc(void *pv, size_t cb);
static void *win32_comtask_realloc(void *pv, size_t cb);
#endif /* WIN32 */

typedef int (*PfnAllocFailHandler)(void *pv, size_t cb, cstring_flags_t flags, void *param);

static void *cstring_realloc_2( void                *pv
                            ,   size_t              cb
                            ,   cstring_flags_t     flags
                            ,   CSTRING_RC          *prc
                            ,   PfnAllocFailHandler	pfnAllocFailHandler
                            ,   void                *param)
{
    void    *pvNew;

    if(CSTRING_F_MEMORY_IS_OFFSET & flags)
    {
        if(NULL != pv)
        {
            pv = (char*)pv - CSTRING_OFFSET_SIZE;

            if(NULL == pfnAllocFailHandler)
            {
                memcpy(&pfnAllocFailHandler, pv, sizeof(pfnAllocFailHandler));
            }
        }

        if(0 != cb)
        {
            cb += CSTRING_OFFSET_SIZE;
        }
    }

alloc_retry:

    switch(CSTRING_F_ARENA_MASK & flags)
    {
        case    CSTRING_F_USE_REALLOC:
#if defined(_MSC_VER) && \
    defined(_DEBUG)
            if(0 == cb)
            {
                free(pv);
                return NULL;
            }
#endif /* _DEBUG */
            pvNew = realloc(pv, cb);
            break;
#if defined(WIN32)
        case    CSTRING_F_USE_WIN32_GLOBAL_MEMORY:
            pvNew = win32_global_realloc(pv, cb);
            break;
        case    CSTRING_F_USE_WIN32_PROCESSHEAP_MEMORY:
            pvNew = win32_processheap_realloc(pv, cb);
            break;
        case    CSTRING_F_USE_WIN32_COM_TASK_MEMORY:
            pvNew = win32_comtask_realloc(pv, cb);
            break;
#endif /* WIN32 */
#if defined(CSTRING_USE_SYNESIS_APIS)
        case    CSTRING_F_USE_SYNESIS_HATOR:
#endif /* CSTRING_USE_SYNESIS_APIS */
        case    CSTRING_F_USE_CUSTOMARENAFUNCTIONS:
        default:
            *prc = CSTRING_RC_INVALIDARENA;
            return NULL;
    }

    if(NULL == pvNew)
    {
        if( 0 != cb &&
            NULL != pfnAllocFailHandler && 
            1 == (*pfnAllocFailHandler)(pv, cb, flags, param))
        {
            goto alloc_retry;
        }

        *prc = CSTRING_RC_OUTOFMEMORY;
    }

    pv = pvNew;

    if(CSTRING_F_MEMORY_IS_OFFSET & flags)
    {
        if(NULL != pv)
        {
            if(NULL != pfnAllocFailHandler)
            {
                memcpy(pv, &pfnAllocFailHandler, sizeof(pfnAllocFailHandler));
            }

            pv = (char*)pv + CSTRING_OFFSET_SIZE;
        }
    }

    return pv;
}

static void *cstring_realloc_(void *pv, size_t cb, cstring_flags_t flags, CSTRING_RC *prc)
{
    return cstring_realloc_2(pv, cb, flags, prc, NULL, NULL);
}

/* /////////////////////////////////////////////////////////////////////////////
 * API
 */

#if defined(_DLL) || \
    defined(__DLL) || \
    1
CSTRING_EXTERN_C CSTRING_RC cstring_apiInit(void);
CSTRING_EXTERN_C void cstring_apiUninit(void);

CSTRING_EXTERN_C CSTRING_RC cstring_apiInit(void)
{
    return CSTRING_RC_SUCCESS;
}

CSTRING_EXTERN_C void cstring_apiUninit(void)
{
}
#endif /* _DLL */

CSTRING_EXTERN_C char const *cstring_error(CSTRING_RC rc)
{
    switch(rc)
    {
        default:    return "<<unknown error>>";
        case    CSTRING_RC_SUCCESS                      : return "Operation completed successfully";
        case    CSTRING_RC_OUTOFMEMORY                  : return "Out of memory";
        case    CSTRING_RC_FIXED                        : return "Operation cannot procede because the cstring has fixed capacity";
        case    CSTRING_RC_BORROWED                     : return "Operation cannot procede because the cstring is borrowed";
        case    CSTRING_RC_READONLY                     : return "Operation cannot procede because the cstring is readonly";
        case    CSTRING_RC_INVALIDARENA                 : return "Invalid arena";
        case    CSTRING_RC_CUSTOMARENANOTSUPPORTED      : return "Custom arena functionality not currently supported";
        case    CSTRING_RC_EXCEEDFIXEDCAPACITY          : return "Operation cannot procede because the current capacity would be exceeded and the cstring has fixed capacity";
        case    CSTRING_RC_EXCEEDBORROWEDCAPACITY       : return "Operation cannot procede because the current capacity would be exceeded and the cstring is borrowed";
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_init(struct cstring_t  *pcs)
{
    cstring_assert(NULL != pcs);

    pcs->len        =   0;
    pcs->ptr        =   NULL;
    pcs->capacity   =   0;
    pcs->flags      =   0;

    return CSTRING_RC_SUCCESS;
}

CSTRING_EXTERN_C CSTRING_RC cstring_create( struct cstring_t    *pcs
                                        ,   char const          *s)
{
    const size_t    len =   cstring_strlen_safe_(s);
    size_t          cb;
    CSTRING_RC      rc;

    cstring_assert(NULL != pcs);

    cb  =   1 + len;
    cb  =   (cb + (CSTRING_ALLOC_GRANULARITY - 1)) & ~(CSTRING_ALLOC_GRANULARITY - 1);

    pcs->ptr = (char*)cstring_realloc_(NULL, cb, 0, &rc);

    if(NULL == pcs->ptr)
    {
        return rc;
    }
    else
    {
        pcs->len        =   len;
        (void)strcpy(pcs->ptr, s);
        pcs->capacity   =   cb;
        pcs->flags      =   CSTRING_F_MEMORY_IS_INTERNAL_HEAP | CSTRING_F_USE_REALLOC;

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_createLen(  struct cstring_t    *pcs
                                        ,       char const          *s
                                        ,       size_t              len)
{
    size_t      cb;
    CSTRING_RC  rc;

    cstring_assert(NULL != pcs);

    cb  =   1 + len;
    cb  =   (cb + (CSTRING_ALLOC_GRANULARITY - 1)) & ~(CSTRING_ALLOC_GRANULARITY - 1);

    pcs->ptr = (char*)cstring_realloc_(NULL, cb, 0, &rc);

    if(NULL == pcs->ptr)
    {
        return rc;
    }
    else
    {
        pcs->len        =   len;
        (void)strncpy(pcs->ptr, s, len);
        pcs->ptr[len]   =   '\0';
        pcs->capacity   =   cb;
        pcs->flags      =   CSTRING_F_MEMORY_IS_INTERNAL_HEAP | CSTRING_F_USE_REALLOC;

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_createEx(   struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       cstring_flags_t         flags
                                        ,       void                    *arena
                                        ,       size_t                  capacity)
{
    const size_t    len =   cstring_strlen_safe_(s);

    return cstring_createLenEx(pcs, s, len, flags, arena, capacity);
}

CSTRING_EXTERN_C CSTRING_RC cstring_createLenEx(struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       size_t                  len
                                        ,       cstring_flags_t         flags
                                        ,       void                    *arena
                                        ,       size_t                  capacity)
{
    return cstring_createLenFn(pcs, s, len, flags, arena, capacity, NULL, NULL);
}

/** This is an experimental feature */
CSTRING_EXTERN_C CSTRING_RC cstring_createLenFn(struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       size_t                  len
                                        ,       cstring_flags_t         flags
                                        ,       void                    *arena
                                        ,       size_t                  capacity
                                        ,       void                    *pfnAllocFailHandler
                                        ,       void                    *param)
{
    cstring_assert(NULL != pcs);

    if(NULL != pfnAllocFailHandler)
    {
        flags |= CSTRING_F_MEMORY_IS_OFFSET;
    }

    if(CSTRING_F_MEMORY_IS_BORROWED & flags)
    {
        flags |= CSTRING_F_MEMORY_IS_FIXED;
    }

    if( NULL != arena &&
        !(CSTRING_F_MEMORY_IS_BORROWED & flags))
    {
        return CSTRING_RC_CUSTOMARENANOTSUPPORTED;
    }
    else
    {
        if(CSTRING_F_MEMORY_IS_BORROWED & flags)
        {
            if(len + 1 >= capacity)
            {
                return CSTRING_RC_EXCEEDBORROWEDCAPACITY;
            }

            pcs->ptr        =   (char*)arena;
            pcs->capacity   =   capacity;
        }
        else
        {
            size_t      cb;
            CSTRING_RC  rc;

            cb  =   1 + len;
            if(cb < capacity)
            {
                cb = capacity;
            }
            cb  =   (cb + (CSTRING_ALLOC_GRANULARITY - 1)) & ~(CSTRING_ALLOC_GRANULARITY - 1);

            pcs->ptr = (char*)cstring_realloc_2(NULL, cb, flags, &rc, (PfnAllocFailHandler)pfnAllocFailHandler, param);

            if(NULL == pcs->ptr)
            {
                return rc;
            }

            pcs->capacity   =   cb;
        }

        pcs->len        =   len;
        (void)strncpy(pcs->ptr, s, len);
        pcs->ptr[len]   =   '\0';
        pcs->flags      =   flags;

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_destroy(struct cstring_t *pcs)
{
    CSTRING_RC  rc  =   CSTRING_RC_SUCCESS;

    cstring_assert(NULL != pcs);

    if(!(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags))
    {
        (void)cstring_realloc_(pcs->ptr, 0, pcs->flags, &rc);
    }

    pcs->len        =   0;
    pcs->ptr        =   NULL;
    pcs->capacity   =   0;
    pcs->flags      =   0;

    return rc;
}

CSTRING_EXTERN_C CSTRING_RC cstring_yield(  struct cstring_t    *pcs
                                        ,   char                **ppBuff)
{
    cstring_assert(NULL != pcs);
    cstring_assert(NULL != ppBuff);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else if(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags)
    {
        return CSTRING_RC_BORROWED;
    }
    else
    {
        *ppBuff = pcs->ptr;

        pcs->ptr        =   NULL;
        pcs->len        =   0;
        pcs->capacity   =   0;

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_yield2(     struct cstring_t        *pcs
                                        ,       char                    **pPayload
                                        ,       void                    **pRaw)
{
    cstring_assert(NULL != pcs);
    cstring_assert(NULL != pPayload);
    cstring_assert(NULL != pRaw);
    cstring_assert((void**)pPayload != pRaw);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else if(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags)
    {
        return CSTRING_RC_BORROWED;
    }
    else
    {
        *pPayload   =   pcs->ptr;
        *pRaw       =   pcs->ptr;

        pcs->ptr        =   NULL;
        pcs->len        =   0;
        pcs->capacity   =   0;

        return CSTRING_RC_SUCCESS;
    }
}


CSTRING_EXTERN_C CSTRING_RC cstring_setCapacity(struct cstring_t    *pcs
                                        ,       size_t              capacity)
{
    return cstring_setCapacityFn(pcs, capacity, NULL, NULL);
}

/** This is an experimental feature */
CSTRING_EXTERN_C CSTRING_RC cstring_setCapacityFn(struct cstring_t        *pcs
                                        ,       size_t                  capacity
                                        ,       void                    *pfnAllocFailHandler
                                        ,       void                    *param)
{
    cstring_assert(NULL != pcs);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else
    {
        if(CSTRING_F_MEMORY_IS_FIXED & pcs->flags)
        {
            return CSTRING_RC_EXCEEDFIXEDCAPACITY;
        }
        else if(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags)
        {
            return CSTRING_RC_EXCEEDBORROWEDCAPACITY;
        }
        else if(capacity != pcs->capacity &&
                capacity > pcs->len)
        {
            CSTRING_RC  rc;
            char        *newPtr =   (char*)cstring_realloc_2(pcs->ptr, capacity, pcs->flags, &rc, (PfnAllocFailHandler)pfnAllocFailHandler, param);

            if(NULL == newPtr)
            {
                return rc;
            }
            else
            {
                pcs->ptr            =   newPtr;
                pcs->ptr[pcs->len]  =   '\0';
            }

            pcs->capacity   =   capacity;
        }

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_assign( struct cstring_t    *pcs
                                        ,   char const          *s)
{
    return cstring_assignFn(pcs, s, NULL, NULL);
}

/** This is an experimental feature */
CSTRING_EXTERN_C CSTRING_RC cstring_assignFn(   struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       void                    *pfnAllocFailHandler
                                        ,       void                    *param)
{
    cstring_assert(NULL != pcs);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else
    {
        const size_t    len =   cstring_strlen_safe_(s);
        size_t          cb;

        cb  =   1 + len;
        cb  =   (cb + (CSTRING_ALLOC_GRANULARITY - 1)) & ~(CSTRING_ALLOC_GRANULARITY - 1);

        if(pcs->capacity < cb)
        {
            if(cb < pcs->capacity * 2)
            {
                cb = pcs->capacity * 2;
            }

            if(CSTRING_F_MEMORY_IS_FIXED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDFIXEDCAPACITY;
            }
            else if(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDBORROWEDCAPACITY;
            }
            else
            {
                CSTRING_RC  rc;
                char        *newPtr =   (char*)cstring_realloc_2(pcs->ptr, cb, pcs->flags, &rc, (PfnAllocFailHandler)pfnAllocFailHandler, param);

                if(NULL == newPtr)
                {
                    return rc;
                }
                else
                {
                    pcs->ptr = newPtr;
                }

                pcs->capacity   =   cb;
            }
        }

        pcs->len = len;
        (void)strcpy(pcs->ptr, s);

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_assignLen(  struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       size_t                  len)
{
    return cstring_assignLenFn(pcs, s, len, NULL, NULL);
}

/** This is an experimental feature */
CSTRING_EXTERN_C CSTRING_RC cstring_assignLenFn(struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       size_t                  len
                                        ,       void                    *pfnAllocFailHandler
                                        ,       void                    *param)
{
    cstring_assert(NULL != pcs);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else
    {
        size_t  cb;

        cb  =   1 + len;
        cb  =   (cb + (CSTRING_ALLOC_GRANULARITY - 1)) & ~(CSTRING_ALLOC_GRANULARITY - 1);

        if(pcs->capacity < cb)
        {
            if(cb < pcs->capacity * 2)
            {
                cb = pcs->capacity * 2;
            }

            if(CSTRING_F_MEMORY_IS_FIXED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDFIXEDCAPACITY;
            }
            else if(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDBORROWEDCAPACITY;
            }
            else
            {
                CSTRING_RC  rc;
                char        *newPtr =   (char*)cstring_realloc_2(pcs->ptr, cb, pcs->flags, &rc, (PfnAllocFailHandler)pfnAllocFailHandler, param);

                if(NULL == newPtr)
                {
                    return rc;
                }
                else
                {
                    pcs->ptr = newPtr;
                }

                pcs->capacity   =   cb;
            }
        }

        pcs->len = len;
        (void)strncpy(pcs->ptr, s, len);
        pcs->ptr[len] = '\0';

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_copy(   struct cstring_t          *pcs
                                        ,   struct cstring_t const    *pcsSrc)
{
    cstring_assert(NULL != pcs);
    cstring_assert(NULL != pcsSrc);

    return cstring_assignLen(pcs, pcsSrc->ptr, pcsSrc->len);
}

CSTRING_EXTERN_C CSTRING_RC cstring_append( struct cstring_t    *pcs
                                        ,   char const          *s)
{
    return cstring_appendFn(pcs, s,NULL, NULL);
}

/** This is an experimental feature */
CSTRING_EXTERN_C CSTRING_RC cstring_appendFn(   struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       void                    *pfnAllocFailHandler
                                        ,       void                    *param)
{
    cstring_assert(NULL != pcs);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else
    {
        const size_t    len =   cstring_strlen_safe_(s);
        size_t          cb;

        cb  =   1 + pcs->len + len;
        cb  =   (cb + (CSTRING_ALLOC_GRANULARITY - 1)) & ~(CSTRING_ALLOC_GRANULARITY - 1);

        if(pcs->capacity < cb)
        {
            if(cb < pcs->capacity * 2)
            {
                cb = pcs->capacity * 2;
            }

            if(CSTRING_F_MEMORY_IS_FIXED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDFIXEDCAPACITY;
            }
            else if(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDBORROWEDCAPACITY;
            }
            else
            {
                CSTRING_RC  rc;
                char        *newPtr =   (char*)cstring_realloc_2(pcs->ptr, cb, pcs->flags, &rc, (PfnAllocFailHandler)pfnAllocFailHandler, param);

                if(NULL == newPtr)
                {
                    return rc;
                }
                else
                {
                    pcs->ptr = newPtr;
                }

                pcs->capacity   =   cb;
            }
        }

        (void)strcpy(pcs->ptr + pcs->len, s);
        pcs->len += len;

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_appendLen(  struct cstring_t    *pcs
                                        ,       char const          *s
                                        ,       size_t              len)
{
    return cstring_appendLenFn(pcs, s, len, NULL, NULL);
}

/** This is an experimental feature */
CSTRING_EXTERN_C CSTRING_RC cstring_appendLenFn(struct cstring_t        *pcs
                                        ,       char const              *s
                                        ,       size_t                  len
                                        ,       void                    *pfnAllocFailHandler
                                        ,       void                    *param)
{
    cstring_assert(NULL != pcs);
    cstring_assert(NULL != s || 0 == len);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else
    {
        size_t  cb;

        cb  =   1 + pcs->len + len;
        cb  =   (cb + (CSTRING_ALLOC_GRANULARITY - 1)) & ~(CSTRING_ALLOC_GRANULARITY - 1);

        if(pcs->capacity < cb)
        {
            if(cb < pcs->capacity * 2)
            {
                cb = pcs->capacity * 2;
            }

            if(CSTRING_F_MEMORY_IS_FIXED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDFIXEDCAPACITY;
            }
            else if(CSTRING_F_MEMORY_IS_BORROWED & pcs->flags)
            {
                return CSTRING_RC_EXCEEDBORROWEDCAPACITY;
            }
            else
            {
                CSTRING_RC  rc;
                char        *newPtr =   (char*)cstring_realloc_2(pcs->ptr, cb, pcs->flags, &rc, (PfnAllocFailHandler)pfnAllocFailHandler, param);

                if(NULL == newPtr)
                {
                    return rc;
                }
                else
                {
                    pcs->ptr = newPtr;
                }

                pcs->capacity   =   cb;
            }
        }

        (void)strncpy(pcs->ptr + pcs->len, s, len);
        pcs->len            +=  len;
        pcs->ptr[pcs->len]  =   '\0';

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_truncate(   struct cstring_t    *pcs
                                        ,       size_t              len)
{
    cstring_assert(NULL != pcs);

    if(CSTRING_F_MEMORY_IS_READONLY & pcs->flags)
    {
        return CSTRING_RC_READONLY;
    }
    else
    {
        if(len < pcs->len)
        {
            pcs->ptr[pcs->len = len] = '\0';
        }

        return CSTRING_RC_SUCCESS;
    }
}

CSTRING_EXTERN_C CSTRING_RC cstring_swap(   struct cstring_t  *pcs1
                                        ,   struct cstring_t  *pcs2)
{
    cstring_assert(NULL != pcs1);
    cstring_assert(NULL != pcs2);

    if( (CSTRING_F_MEMORY_IS_READONLY & pcs1->flags) ||
        (CSTRING_F_MEMORY_IS_READONLY & pcs2->flags))
    {
        return CSTRING_RC_READONLY;
    }
    else if((CSTRING_F_MEMORY_IS_BORROWED & pcs1->flags) ||
            (CSTRING_F_MEMORY_IS_BORROWED & pcs2->flags))
    {
        return CSTRING_RC_BORROWED;
    }
    else
    {
        struct cstring_t  cs;

        cs      =   *pcs1;
        *pcs1   =   *pcs2;
        *pcs2   =   cs;

        return CSTRING_RC_SUCCESS;
    }
}

/* ////////////////////////////////////////////////////////////////////////// */

#if defined(WIN32)

# ifndef CSTRING_DOCUMENTATION_SKIP_SECTION
#  if defined(__MWERKS__)
#   define  GMEM_FIXED      0
#   define  GMEM_MOVEABLE   2
#   define  WINAPI          __stdcall
typedef int             BOOL;
typedef void            *HANDLE;
typedef void            *HGLOBAL;
typedef void            *HMODULE;
typedef void            *HINSTANCE;
typedef long            LONG;
__declspec(dllimport) long      __stdcall   InterlockedExchange(LONG volatile *, LONG);
__declspec(dllimport) HANDLE    __stdcall   GetProcessHeap(void);
__declspec(dllimport) HGLOBAL   __stdcall   GlobalAlloc(unsigned int, unsigned long);
__declspec(dllimport) HGLOBAL   __stdcall   GlobalFree(HGLOBAL);
__declspec(dllimport) HGLOBAL   __stdcall   GlobalReAlloc(HGLOBAL, unsigned long, unsigned int);
__declspec(dllimport) void      *__stdcall  HeapAlloc(HANDLE, unsigned long, unsigned int);
__declspec(dllimport) BOOL      __stdcall   HeapFree(HANDLE, unsigned long, void *);
__declspec(dllimport) void      *__stdcall  HeapReAlloc(HANDLE, unsigned long, void*, unsigned int);
__declspec(dllimport) void      __stdcall   Sleep(unsigned long);
__declspec(dllimport) HINSTANCE __stdcall   LoadLibraryA(char const *);
__declspec(dllimport) BOOL      __stdcall   FreeLibrary(HMODULE );
__declspec(dllimport) void      *__stdcall  GetProcAddress(HMODULE, char const*);
#  else /* ? compiler */
#   include <windows.h>
#  endif /* compiler */
# endif /* !CSTRING_DOCUMENTATION_SKIP_SECTION */

/* ////////////////////////////////////////////////////////////////////////// */

static void *win32_global_realloc(void *pv, size_t cb)
{
    /* Logic borrowed from implementation of SynesisWin::GlobalAtor class (in
     * file MWAtors.h) from the Synesis Software Public Domain Source Code 
     * Library (http://synesis.com.au/software).
     */
    if(NULL != pv)
    {
        if(0 == cb)
        {
            return (GlobalFree((HGLOBAL)pv), (void*)NULL);
        }
        else
        {
            return (void*)GlobalReAlloc((HGLOBAL)pv, cb, GMEM_MOVEABLE);
        }
    }
    else
    {
        return (void*)GlobalAlloc(GMEM_FIXED, cb);
    }
}

static void *win32_processheap_realloc(void *pv, size_t cb)
{
    /* Logic borrowed from implementation of SynesisWin::HeapAtor class (in
     * file MWAtors.h) from the Synesis Software Public Domain Source Code 
     * Library (http://synesis.com.au/software).
     */
    if(NULL != pv)
    {
        if(0 == cb)
        {
            return (HeapFree(GetProcessHeap(), 0, (HGLOBAL)pv), (void*)NULL);
        }
        else
        {
            return (void*)HeapReAlloc(GetProcessHeap(), 0, (HGLOBAL)pv, cb);
        }
    }
    else
    {
        return (void*)HeapAlloc(GetProcessHeap(), 0, cb);
    }
}

static void *win32_comtask_realloc(void *pv, size_t cb)
{
    typedef void* (WINAPI *PfnCoTaskMemRealloc)(void *, size_t );

    static LONG                 s_cstring_PfnCoTaskMemRealloc_init  =   0;
    static LONG                 s_cstring_PfnCoTaskMemRealloc_spin  =   0;
    static HINSTANCE            s_cstring_ole32_HINSTANCE           =   NULL;
    static PfnCoTaskMemRealloc  s_cstring_pfnCoTaskMemRealloc       =   NULL;

    if(NULL == pv)
    {
        for(; 0 != InterlockedExchange(&s_cstring_PfnCoTaskMemRealloc_spin, 1); Sleep(0))
        {}

        if(1 == ++s_cstring_PfnCoTaskMemRealloc_init)
        {
            s_cstring_ole32_HINSTANCE       =   LoadLibraryA("OLE32");
            s_cstring_pfnCoTaskMemRealloc   =   (PfnCoTaskMemRealloc)GetProcAddress(s_cstring_ole32_HINSTANCE, "CoTaskMemRealloc");

            if( NULL == s_cstring_ole32_HINSTANCE ||
                NULL == s_cstring_pfnCoTaskMemRealloc)
            {
                --s_cstring_PfnCoTaskMemRealloc_init;
            }
        }

        InterlockedExchange(&s_cstring_PfnCoTaskMemRealloc_spin, 0);
    }

    pv = (NULL != s_cstring_pfnCoTaskMemRealloc) ? s_cstring_pfnCoTaskMemRealloc(pv, cb) : NULL;

    if(0 == cb)
    {
        for(; 0 != InterlockedExchange(&s_cstring_PfnCoTaskMemRealloc_spin, 1); Sleep(0))
        {}

        if(0 == --s_cstring_PfnCoTaskMemRealloc_init)
        {
            FreeLibrary(s_cstring_ole32_HINSTANCE);
        }

        InterlockedExchange(&s_cstring_PfnCoTaskMemRealloc_spin, 0);
    }

    return pv;
}

/* ////////////////////////////////////////////////////////////////////////// */

#endif /* WIN32 */

/* /////////////////////////////////////////////////////////////////////////////
 * Documentation
 */

/** \mainpage What is cstring?
 *
 * \section section__introduction Introduction
 *
 * <b>cstring</b> is a small and simple C library for the definition and manipulation
 * of expandable C-style strings.
 *
 * \section section__license License
 *
 * <b>cstring</b> is released under the <a href = "http://www.opensource.org/licenses/bsd-license.html">BSD</a>
 * license, which basically means its free for any use, but you can't claim it's yours.
 *
 * \section section__c_api C-API
 *
 * The library is implemented in C, and provides a simple C-API, consisting of
 * the following functions:
 * \li cstring_error()
 * \li cstring_init()
 * \li cstring_create()
 * \li cstring_createLen()
 * \li cstring_createEx()
 * \li cstring_createLenEx()
 * \li cstring_destroy()
 * \li cstring_yield()
 * \li cstring_setCapacity()
 * \li cstring_assign()
 * \li cstring_assignLen()
 * \li cstring_copy()
 * \li cstring_append()
 * \li cstring_appendLen()
 * \li cstring_truncate()
 * \li cstring_swap()
 *
 * Consider the following example, which creates a cstring instance, and then
 * appends two other C-style strings to it, prints it to stdout, truncates it,
 * prints it again, and then destroys it:
 *
 * \code
  cstring_t       str;

  // create a string from the C-style string "Hello,"
  cstring_create(&str, "Hello,");

  // append a space
  cstring_appendLen(&str, " ", 1);

  // append "World!"
  cstring_append(&str, "World!");

  // print it out, passing length+ptr to printf
  printf("cstring contents: %.*s\n", str.len, str.ptr);

  // truncate it to length 5
  cstring_truncate(&str, 5);

  // print it again, passing the ptr member to puts()
  puts(str.ptr);

  // destroy it
  cstring_destroy(&str);
 * \endcode
 *
 *
 * \subsection section__c_api__custom_memory Custom Memory
 *
 * <b>cstring</b> supports different allocation from differnet memory libraries.
 * For example, to create a string that allocates its buffer from the Win32
 * process heap, specify the CSTRING_F_USE_WIN32_PROCESSHEAP_MEMORY to
 * cstring_createEx, as in:
 *
 * \code
  cstring_t       str;

  cstring_createEx(&str, "G'day!", CSTRING_F_USE_WIN32_PROCESSHEAP_MEMORY, NULL, 0);
 * \endcode
 *
 * All subsequent operations that require memory (de-)allocation on this string
 * will use thw Win32 process heap, e.g.:
 *
 * \code

  cstring_assign(&str, "Welcome to the Pleasuredome"); // Will go to the Win32 process heap
 * \endcode
 *
 * \subsection section__c_api__borrowed_memory Borrowed Memory
 *
 * <b>cstring</b> supports the specification of external memory blocks for use
 * by cstring instances, via specification of the CSTRING_F_MEMORY_IS_BORROWED
 * flag, as in:
 *
 * \code
  char                buff[101];
  struct cstring_t    str;
  CSTRING_RC          rc  =   cstring_createEx( &str
                                              , "Some contents for this "
                                              , CSTRING_F_MEMORY_IS_BORROWED
                                              , &buff[0]
                                              , sizeof(buff));

 * \endcode
 * The \c str instance will do all its memory (de-)allocation with respect to
 * the fixed buffer \c buff. It will not comply with any operation that
 * requires it to have a capacity of more than \c sizeof(buff).
 */

/* ////////////////////////////////////////////////////////////////////////// */
